// ========================================================================
//
// emudbg.c
//
// A small collection of functions to make debugging easier when running
// under the Helio Emulator
//
// Copyright (c) 2000 Frank Somers, Ground Zero Development
// <frank@g0dev.com>
//
// This program is free software; you can redistribute it and/or modify it
// provided the copyright messages are retained in the source code.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE
//
// NOTE: This is a component that is shared between the emulator source
//       and applications running under the emulator that wish to use
//       the debugging services.  You are allowed to freely use this file
//       in your application without any restrictions other than those
//       stated above.
//
// ========================================================================

#include "emudbg.h"

#ifdef   EMUDBG_ENABLED
   #include "include\mmu.h"

   #ifndef  __GNUC__
      #error "This code depends on GNU C extensions"
   #endif


   // ------------------------------------------------------------------------
   //
   // Private data
   //
   //

   // This will be set by DbgAppBegin () if it detects a compatible
   // Emulator coprocessor

   static int s_fEnabled = 0;


   // Provide GDB with a temporary location it can use when executing
   // functions in the target, eg 'p Func ()'
   // If this isn't provided, it will assume it can use the programs
   // start addr, which isn't posible if we're running in ROM

   unsigned long __CALL_DUMMY_ADDRESS = 0;


   // ------------------------------------------------------------------------
   //
   // Internal utils for dealing with the system coprocessor
   //
   // NOTE: GAS will insert the necessary delay slot filler instructions
   //

   static __inline UWORD GetCop0Status (void) {
      UWORD status;
      asm ("mfc0 %0, $12" : "=r" (status));
      return status;
   }

   static __inline void SetCop0Status (UWORD status) {
      asm ("mtc0 %0, $12" : : "r" (status));
   }

   static __inline UWORD EnableCop2 (void) {
      SetCop0Status (GetCop0Status () | 0x40000000);
   }

   static __inline UWORD DisableCop2 (void) {
      SetCop0Status (GetCop0Status () & ~0x40000000);
   }


   // ------------------------------------------------------------------------
   //
   // Internal utils for dealing with the emu coprocessor
   //
   // NOTE: GAS will insert the necessary delay slot filler instructions
   //

   // Emulator coprocessor revision numbers.  The major version must match
   // or we can't use it.  If the major ver matches, the minor version must
   // be greater or equal to what we require

   #define  EMUCOP_SIG_VALUE           0x454d5500  // 'EMU'
   #define  EMUCOP_SIG_MASK            0xffffff00
   #define  EMUCOP_VER_MAJOR           0
   #define  EMUCOP_VER_MINOR           0


   // Command codes for EmuCommand

   #define  EMUCOP_CMD_AUXTLB_RESET    0
   #define  EMUCOP_CMD_GDB_BREAK       1



   static __inline UWORD EmuGetVersion (void) {
      UWORD val;
      asm ("mfc2 %0, $0" : "=r" (val));
      return val;
   }

   static __inline UWORD EmuGetStatus (void) {
      UWORD val;
      asm ("mfc2 %0, $2" : "=r" (val));
      return val;
   }

   static __inline void EmuCommand (const UWORD cmd) {
      asm ("mtc2 %0, $3" : : "r" (cmd));
   }

   static __inline UWORD EmuGetAuxTlbSize (void) {
      UWORD val;
      asm ("mfc2 %0, $4" : "=r" (val));
      return val;
   }

   static __inline UWORD EmuGetAuxTlbIndex (void) {
      UWORD val;
      asm ("mfc2 %0, $5" : "=r" (val));
      return val;
   }

   static __inline void EmuSetAuxTlbIndex (const UWORD idx) {
      asm ("mtc2 %0, $5" : : "r" (idx));
   }

   static __inline UWORD EmuGetAuxTlbEntryHi (void) {
      UWORD val;
      asm ("mfc2 %0, $6" : "=r" (val));
      return val;
   }

   static __inline void EmuSetAuxTlbEntryHi (const UWORD entryhi) {
      asm ("mtc2 %0, $6" : : "r" (entryhi));
   }

   static __inline UWORD EmuGetAuxTlbEntryLo (void) {
      UWORD val;
      asm ("mfc2 %0, $7" : "=r" (val));
      return val;
   }

   static __inline void EmuSetAuxTlbEntryLo (const UWORD entrylo) {
      asm ("mtc2 %0, $7" : : "r" (entrylo));
   }

   static __inline void EmuPutc (const UWORD c) {
      asm ("mtc2 %0, $8" : : "r" (c));
   }


   static int CheckForEmu (void) {

      // Check for emulator extension coprocessor.  Returns != 0 if
      // its compatible.  Assumes cop2 already enabled or this will fault

      UWORD ver = EmuGetVersion ();
      if ((ver & EMUCOP_SIG_MASK) != EMUCOP_SIG_VALUE)
         return 0;

      // Low byte has a major/ minor version in it.  Check for compatibility

      if (((ver >> 4) & 0xf) != EMUCOP_VER_MAJOR)
         return 0;   // Incompatible major version

   #if   (EMUCOP_VER_MINOR != 0)
      if ((ver & 0x0f) < EMUCOP_VER_MINOR)
         return 0;   // We require a higher minor version
   #endif

      // All looks OK

      return 1;
   }


   static void AddAuxTlbEntries (UWORD vaddr, UWORD iTlb,
      UWORD cEntries, UWORD * pulPages) {

      // This prepares the auxiliary TLB entries as the O/S does during a
      // TLB exception

      int i;
      for (i = 0; i < cEntries; vaddr += BLOCK_SIZE, i++) {
         EmuSetAuxTlbIndex (iTlb + i);
         EmuSetAuxTlbEntryHi (vaddr);
         EmuSetAuxTlbEntryLo ((pulPages [i] & 0x1fffffff) | 0x0700);
      }
   }


   // ------------------------------------------------------------------------
   //
   // Registering and unregistering the app
   //
   // (badvpn << 12), (phys & 0x1fffffff) | 0x700
   //

   void DbgAppBegin (void) {

      // Enable Cop2 and check for a compatible EMU coprocessor.  We may be
      // running on the emulator or on the real device.

      s_fEnabled = 0;
      EnableCop2 ();
      if (!CheckForEmu ()) {
         DisableCop2 ();
         return;
      }
      s_fEnabled = 1;

      // Load up the initial virtual memory state

      DbgRefreshMemInfo ();
   }


   void DbgRefreshMemInfo (void) {

      // Propagate information about the virual memory configuration to
      // the emulator so it can be used during debugging.  If this isn't
      // done, then GDB will only be able to access memory that is
      // currently mapped in the TLB

      UWORD iTlb;
      UWORD * pulTable;

      if (s_fEnabled) {
         // Reset the auxiliary TLB

         EmuCommand (EMUCOP_CMD_AUXTLB_RESET);
         iTlb = 0;

         // Add the app TLB entries

         MemoryGetPageTable (0, &pulTable, 0);
         AddAuxTlbEntries (USER_MEM_BASE, iTlb, pulTable [-1], pulTable);
         iTlb += pulTable [-1];

         // Add the system TLB entries

         MemoryGetPageTable (&pulTable, 0, 0);
         AddAuxTlbEntries (SYS_MEM_BASE, iTlb, pulTable [-1], pulTable);
         iTlb += pulTable [-1];

         // Add the interrupt TLB entries

         // MemoryGetPageTable (0, 0, &pulTable);
         // AddAuxTlbEntries (INT_MEM_BASE, iTlb, pulTable [-1], pulTable);
         // iTlb += pulTable [-1];
      }
   }


   void DbgAppEnd (void) {
      if (s_fEnabled)
         EmuCommand (EMUCOP_CMD_AUXTLB_RESET);
      s_fEnabled = 0;
   }


   // ------------------------------------------------------------------------
   //
   // General debugging utils
   //
   //

   void DbgBreak (void) {
      if (s_fEnabled)
         EmuCommand (EMUCOP_CMD_GDB_BREAK);
   }


   void DbgMessage (register const char * pszMsg) {
      if (s_fEnabled) {
         register UWORD c;
         while (c = (UWORD)(unsigned char)*pszMsg++)
            EmuPutc (c);
      }
   }


   // DbgPrintf (format, ...) would have been nice, but vsprintf () isn't
   // exposed by the O/S.  You'll just have to sprintf and DbgMessage instead

#endif   // EMUDBG_ENABLED
