// ========================================================================
//
// emudbg.h
//
// A small collection of functions to make debugging easier when running
// under the Helio Emulator
//
// Copyright (c) 2000 Frank Somers, Ground Zero Development
// <frank@g0dev.com>
//
// This program is free software; you can redistribute it and/or modify it
// provided the copyright messages are retained in the source code.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE
//
// NOTE: This is a component that is shared between the emulator source
//       and applications running under the emulator that wish to use
//       the debugging services.  You are allowed to freely use this file
//       in your application without any restrictions other than those
//       stated above.
//
// ========================================================================

#ifndef  EMUDBG_H
#define  EMUDBG_H

//
// How to use this mini-library...
//
// These provide auxiliary debugger support that can be used while running
// in the Helio emulator.  They will detect if they are running on the
// emulator or not, and will do nothing if not running under the emu.
//
// In your makefile, make sure that -DDEBUG, or -DEMUDBG_ENABLED is on the
// compiler command line when you are doing your builds.  When either of
// these are defined, the functionality in this module will be enabled.
// If not, they will all be compiled out.  You can remote the -Ds when doing
// your release builds to compile out all of the extra code.
//
// In the makefile, add emudbg.o into the list of object files in your app
//
// Include this file in any/ all of your source files as necessary.
//
// DBG_APP_BEGIN()
//    Insert a call to this near the very start of your program.  Typically
//    it will be the first thing that is done in your __main () function
//
// DBG_APP_END()
//    Insert a call to this near the very end of your program.  Typically
//    it will be the last thing that is done in your __main () function
//
// DBG_REFRESH_MEM_INFO()
//    If your application does significant dynamic memory allocation, you
//    may need to include a call to this after your allocation calls.
//    This rebuilds the virtual memory information needed by the debugger.
//    If when using GDB, you notice that you can't examine memory that
//    you have allocated, adding this call may solve the problem
//
// DBG_BREAK()
//    Insert a call to this anywhere where you would like to automatically
//    break into GDB when its executed.  It will only break to GDB if it
//    is currently connected; otherwise execution will continue
//
// DBG_MESSAGE(sz)
//    You can call this at any time to output a message to the emulator
//    Log Window.  You have to format the string yourself (eg with sprintf).
//    Use '\n' for end of line
//
//
// For example, to have your app start up and immediately break to GDB
// (assuming its already connected to the emulator), your __main () function
// would look something like this:
//
//    int __main (WORD cmd, void *cmd_ptr) {
//       DBG_APP_BEGIN ();
//       DBG_BREAK ();
//    	YourAppLaunch (cmd, (void*)cmd_ptr);
//       DBG_APP_END ();
//    }


#if   defined (DEBUG) && !defined (EMUDBG_ENABLED)
   #define  EMUDBG_ENABLED
#endif

#ifdef   EMUDBG_ENABLED
   #ifdef   __cplusplus
      extern "C" {
   #endif

   void DbgAppBegin (void);
   void DbgAppEnd (void);
   void DbgRefreshMemInfo (void);

   void DbgBreak (void);
   void DbgMessage (const char * pszMsg);

   #ifdef   __cplusplus
      }
   #endif

   #define  DBG_APP_BEGIN()         DbgAppBegin ()
   #define  DBG_APP_END()           DbgAppBegin ()
   #define  DBG_REFRESH_MEM_INFO()  DbgRefreshMemInfo ()
   #define  DBG_BREAK()             DbgBreak ()
   #define  DBG_MESSAGE(sz)         DbgMessage (sz)

#else

   #define  DBG_APP_BEGIN()         /* nothing */
   #define  DBG_APP_END()           /* nothing */
   #define  DBG_REFRESH_MEM_INFO()  /* nothing */
   #define  DBG_BREAK()             /* nothing */
   #define  DBG_MESSAGE(sz)         /* nothing */

#endif

#endif
